@@ -13,6 +13,7 @@ import numpy as np
 
 # FIXME: replace os with pathlib where possible
 import os
+import pathlib
 import re
 
 # import scipy.optimize
@@ -28,13 +29,19 @@ __all__ = [
     "ShellCommand",
     "Task",
     "FragItFragmentStructure",
+    "MultiwfnNTO",
+    "MultiwfnTotalESP",
     "MultiwfnVolume",
     "PackmolSolvate",
     #    "ExecuteQChem",
     "QChemAIMD",
     "QChemKoopmanError",
+    "QChemKoopmanErrorLPSCF",
+    "QChemLPSCF",
+    "QChemLPSCFRS",
     "QChemLRTDDFT",
     "QChemMinimizeKoopmanError",
+    "QChemMinimizeKoopmanErrorLPSCF",
     "QChemOptimize",
     "QChemPolarizability",
     #    "QChemRTTDDFT",
@@ -46,6 +53,7 @@ __all__ = [
     "WriteQChemInputPolarizability",
     "WriteQChemInputSinglePoint",
     #    "WriteQChemTDSCF",
+    "XTBOptimize",
 ]
 
 
@@ -121,7 +129,7 @@ class InputTask(Task):
         super().__init__(1, handlers=handlers, name=name)
         self.value = value
 
-    def run(self) -> Any:
+    def run(self, *args) -> Any:
         return self.value
 
 
@@ -305,31 +313,14 @@ class FragItFragmentStructure(ExternalTask):
         self, structure: Union[str, mtr.Structure]
     ) -> Tuple[str, Iterable[mtr.Structure]]:
         with self.io() as io:
-            if isinstance(structure, str):
-                filepath = mtr.expand(structure)
-            else:
-                while isinstance(structure, mtr.Structure):
-                    filepath = mtr.expand(
-                        os.path.join(io.work_dir, f"{uuid.uuid4().hex}.xyz")
-                    )
-                    try:
-                        structure.write(filepath)
-                        break
-                    except FileExistsError:
-                        continue
+            structure.write(io.inp)
 
-            self.engine.execute(
-                structure_filepath=filepath, log_filepath=io.out, work_dir=io.work_dir,
-            )
-
-            name, _ = os.path.splitext(os.path.basename(filepath))
-            pat = re.compile(rf"(?P<file_name>{name}_fragment_\d*\.xyz)")
-            matches = (pat.match(s) for s in os.listdir(io.work_dir))
+            self.engine.execute(self.io)
+            name = pathlib.Path(io.inp).stem
 
             return tuple(
-                mtr.Structure.read(os.path.join(io.work_dir, m.group("file_name")))
-                for m in matches
-                if m is not None
+                mtr.Structure.read(str(p))
+                for p in pathlib.Path(io.work_dir).glob(f"{name}_fragment_*.xyz")
             )
 
 
@@ -344,7 +335,7 @@ class MultiwfnBaseTask(ExternalTask):
         raise NotImplementedError
 
     def run(self, filepath: str) -> Any:
-        inp = mtr.MultiwfnInput(mtr.expand(filepath), *self.commands(), -10,)
+        inp = mtr.MultiwfnInput(mtr.expand(filepath), *self.commands(), -10)
 
         with self.io() as io:
             inp.write(io.inp)
@@ -370,6 +361,79 @@ class MultiwfnBaseTask(ExternalTask):
 #         self.engine.execute(input_path=self.input_path)
 
 
+# excitations = mtr.QChemOutput(filepath=filepath).get("electronic_excitations")
+# excitation_filepath = mtr.expand(f"{io.work_dir}/excitations.txt")
+# with open(excitation_filepath, "w") as f:
+#     f.write(excitations.to_gaussian())
+
+# inp = mtr.MultiwfnInput(mtr.expand(filepath), *self.commands(excitation_filepath), -10)
+# inp.write(io.inp)
+
+
+class MultiwfnNTO(MultiwfnBaseTask):
+    def commands(
+        self, excitation_filepath: str, work_dir: str
+    ) -> Tuple[Union[str, int, float]]:
+        return (
+            18,
+            *(
+                a
+                for i in range(40)
+                for a in [6]
+                + ([excitation_filepath] if i == 0 else [])
+                + [i + 1, 2, mtr.expand(f"{work_dir}/S{i+1}.fch")]
+            ),
+            0,
+        )
+
+    def parse(self, output: str) -> None:
+        # FIXME: is there some useful output to return?
+        return None
+
+    def run(self, filepath: str, excitation_filepath: str) -> Any:
+        with self.io() as io:
+            inp = mtr.MultiwfnInput(
+                mtr.expand(filepath),
+                *self.commands(excitation_filepath, io.work_dir),
+                -10,
+            )
+            inp.write(io.inp)
+
+            self.engine.execute(self.io)
+
+            return self.parse(io.out)
+
+
+class MultiwfnTotalESP(MultiwfnBaseTask):
+    def __init__(
+        self,
+        grid_quality: str,
+        engine: mtr.Engine,
+        io: mtr.IO,
+        handlers: Optional[Iterable[mtr.Handler]] = None,
+        name: Optional[str] = None,
+    ) -> None:
+        self.grid_quality = grid_quality
+        super().__init__(engine=engine, io=io, handlers=handlers, name=name)
+
+    def commands(self):
+        grid_quality = dict(low=1, medium=2, high=3)
+        # FIXME: finish implementing this
+        raise NotImplementedError
+        return (
+            5,
+            12,
+            grid_quality[self.grid_quality],
+            "0,0,0",
+            0,
+        )
+
+    def parse(self, output: str) -> mtr.Qty:
+        # FIXME: finish implementing this
+        raise NotImplementedError
+        return mtr.MultiwfnOutput(output).get("volume")
+
+
 class MultiwfnVolume(MultiwfnBaseTask):
     def commands(self):
         integration_mesh_exp = 9  #: int = 9,
@@ -693,6 +757,7 @@ class QChemBaseTask(ExternalTask):
             molecule=structure
             if isinstance(structure, mtr.Structure)
             or isinstance(structure, mtr.QChemStructure)
+            or isinstance(structure, mtr.QChemFragments)
             else mtr.QChemFragments(structures=structure),
             settings=self.defaults(s),
         )
@@ -742,6 +807,72 @@ class QChemAIMD(QChemBaseTask):
         return settings
 
 
+class QChemLPSCF(QChemBaseTask):
+    def parse(self, output: str) -> Any:
+        try:
+            energy = cclib.io.ccread(output).scfenergies * mtr.eV
+        except AttributeError:
+            energy = None
+
+        return energy
+
+    def defaults(self, settings: mtr.Settings) -> mtr.Settings:
+        if ("rem", "exchange") not in settings and ("rem", "method",) not in settings:
+            settings["rem", "exchange"] = "HF"
+        if ("rem", "basis") not in settings:
+            settings["rem", "basis"] = "3-21G"
+        if ("rem", "jobtype") not in settings:
+            settings["rem", "jobtype"] = "sp"
+        if ("rem", "frgm_method") not in settings:
+            settings["rem", "frgm_method"] = "stoll"
+        if ("rem_frgm", "scf_convergence") not in settings:
+            settings["rem_frgm", "scf_convergence"] = 2
+        if ("rem_frgm", "thresh") not in settings:
+            settings["rem_frgm", "thresh"] = 5
+
+        return settings
+
+    def run(
+        self,
+        *fragments: Union[mtr.QChemStructure, mtr.Structure],
+        settings: Optional[mtr.Settings] = None,
+    ) -> Any:
+        s = mtr.Settings() if settings is None else copy.deepcopy(settings)
+
+        # FIXME: this is essentially a hotpatch to handle fragments - come up with something more elegant/sensible ASAP
+        inp = mtr.QChemInput(
+            molecule=mtr.QChemFragments(structures=fragments),
+            settings=self.defaults(s),
+        )
+
+        with self.io() as io:
+            inp.write(io.inp)
+
+            self.engine.execute(self.io)
+
+            return self.parse(io.out)
+
+
+class QChemLPSCFRS(QChemBaseTask):
+    def defaults(self, settings: mtr.Settings) -> mtr.Settings:
+        if ("rem", "exchange") not in settings and ("rem", "method",) not in settings:
+            settings["rem", "exchange"] = "HF"
+        if ("rem", "basis") not in settings:
+            settings["rem", "basis"] = "3-21G"
+        if ("rem", "jobtype") not in settings:
+            settings["rem", "jobtype"] = "sp"
+        if ("rem", "frgm_method") not in settings:
+            settings["rem", "frgm_method"] = "stoll"
+        if ("rem_frgm", "scf_convergence") not in settings:
+            settings["rem_frgm", "scf_convergence"] = 2
+        if ("rem_frgm", "thresh") not in settings:
+            settings["rem_frgm", "thresh"] = 5
+        if ("rem_frgm", "frgm_lpcorr") not in settings:
+            settings["rem_frgm", "frgm_lpcorr"] = "rs_exact_scf"
+
+        return settings
+
+
 class QChemKoopmanError(Task):
     def __init__(
         self,
@@ -786,41 +917,133 @@ class QChemKoopmanError(Task):
         # if ("structure", "multiplicity") not in settings:
         #     settings["structure", "multiplicity"] = 1
 
-        input_settings = InputTask(self.defaults(s))
+        input_settings = self.defaults(s)
 
-        gs = QChemSinglePointFrontier(self.engine, self.gs_io)
-        gs_structure = InputTask(
-            mtr.QChemStructure(structure, charge=0, multiplicity=1)
-        )
+        gs = QChemSinglePointFrontier(self.engine, self.gs_io, name="gs")
+        gs_structure = mtr.QChemStructure(structure, charge=0, multiplicity=1)
         gs.requires(structure=gs_structure, settings=input_settings)
 
-        cation = QChemSinglePoint(self.engine, self.cation_io)
-        cation_structure = InputTask(
-            mtr.QChemStructure(structure, charge=1, multiplicity=2)
-        )
+        cation = QChemSinglePoint(self.engine, self.cation_io, name="cation")
+        cation_structure = mtr.QChemStructure(structure, charge=1, multiplicity=2)
         cation.requires(structure=cation_structure, settings=input_settings)
 
-        anion = QChemSinglePoint(self.engine, self.anion_io)
-        anion_structure = InputTask(
-            mtr.QChemStructure(structure, charge=-1, multiplicity=2)
-        )
+        anion = QChemSinglePoint(self.engine, self.anion_io, name="anion")
+        anion_structure = mtr.QChemStructure(structure, charge=-1, multiplicity=2)
         anion.requires(structure=anion_structure, settings=input_settings)
 
-        wf = Workflow(
-            gs,
-            cation,
-            anion,
-            input_settings,
-            gs_structure,
-            cation_structure,
-            anion_structure,
-        )
+        wf = Workflow(gs, cation, anion)
 
         out = wf.run(available_cores=self.num_cores, num_consumers=num_consumers)
 
-        energy, homo, lumo = out[0]
-        cation = out[1]
-        anion = out[2]
+        energy, homo, lumo = out["gs"]
+        cation = out["cation"]
+        anion = out["anion"]
+
+        ea = energy - anion
+        ip = cation - energy
+
+        J_squared = (ea + lumo) ** 2 + (ip + homo) ** 2
+
+        return np.sqrt(J_squared.convert(mtr.eV ** 2).value.item())
+
+
+class QChemKoopmanErrorLPSCF(Task):
+    def __init__(
+        self,
+        engine: mtr.Engine,
+        gs_io: mtr.IO,
+        cation_io: mtr.IO,
+        anion_io: mtr.IO,
+        handlers: Optional[Iterable[mtr.Handler]] = None,
+        name: Optional[str] = None,
+    ) -> None:
+        super().__init__(
+            (engine.num_threads or 1) * (engine.num_processors or 1),
+            handlers=handlers,
+            name=name,
+        )
+        self.engine = engine
+        self.gs_io = gs_io
+        self.cation_io = cation_io
+        self.anion_io = anion_io
+
+    def defaults(self, settings: mtr.Settings) -> mtr.Settings:
+        if ("rem", "exchange") not in settings and ("rem", "method") not in settings:
+            settings["rem", "exchange"] = "hf"
+        if ("rem", "basis") not in settings:
+            settings["rem", "basis"] = "3-21G"
+        if ("rem", "jobtype") not in settings:
+            settings["rem", "jobtype"] = "sp"
+
+        return settings
+
+    def run(
+        self,
+        *fragments: mtr.Structure,
+        active_fragment: int,
+        settings: Optional[mtr.Settings] = None,
+        num_consumers: Optional[int] = 1,
+    ) -> float:
+        s = mtr.Settings() if settings is None else copy.deepcopy(settings)
+
+        # if ("structure", "charge") not in settings:
+        #     settings["structure", "charge"] = 0
+        #     # FIXME: rather than guessing 0, use rdkit.Chem.rdmolops.GetFormalCharge?
+        # if ("structure", "multiplicity") not in settings:
+        #     settings["structure", "multiplicity"] = 1
+
+        gs_charges = [0] * len(fragments)
+        gs_multiplicities = [1] * len(fragments)
+
+        cation_charges = gs_charges.copy()
+        cation_multiplicities = gs_multiplicities.copy()
+        cation_charges[active_fragment] += 1
+        cation_multiplicities[active_fragment] = 2
+
+        anion_charges = gs_charges.copy()
+        anion_multiplicities = gs_multiplicities.copy()
+        anion_charges[active_fragment] -= 1
+        anion_multiplicities[active_fragment] = 2
+
+        input_settings = self.defaults(s)
+
+        gs = QChemSinglePointFrontier(self.engine, self.gs_io, name="gs")
+        gs_structure = mtr.QChemFragments(
+            fragments,
+            gs_charges,
+            gs_multiplicities,
+            total_charge=sum(gs_charges),
+            total_multiplicity=1,
+        )
+        gs.requires(structure=gs_structure, settings=input_settings)
+
+        cation = QChemSinglePoint(self.engine, self.cation_io, name="cation")
+        cation_structure = mtr.QChemFragments(
+            fragments,
+            cation_charges,
+            cation_multiplicities,
+            total_charge=sum(cation_charges),
+            total_multiplicity=2,
+        )
+        cation.requires(structure=cation_structure, settings=input_settings)
+
+        anion = QChemSinglePoint(self.engine, self.anion_io, name="anion")
+        anion_structure = mtr.QChemFragments(
+            fragments,
+            anion_charges,
+            anion_multiplicities,
+            total_charge=sum(anion_charges),
+            total_multiplicity=2,
+        )
+        anion.requires(structure=anion_structure, settings=input_settings)
+
+        wf = Workflow(gs, cation, anion)
+
+        out = wf.run(available_cores=self.num_cores, num_consumers=num_consumers)
+
+        energy, homo, lumo = out["gs"]
+        cation = out["cation"]
+        anion = out["anion"]
 
         ea = energy - anion
         ip = cation - energy
@@ -832,8 +1055,7 @@ class QChemKoopmanError(Task):
 
 class QChemLRTDDFT(QChemBaseTask):
     def parse(self, output: str) -> Any:
-        # FIXME: implement using cclib
-        raise NotImplementedError
+        return mtr.QChemOutput(filepath=output).get("electronic_excitations")
 
     def defaults(self, settings: mtr.Settings) -> mtr.Settings:
         if ("rem", "exchange") not in settings and ("rem", "method",) not in settings:
@@ -887,6 +1109,81 @@ class QChemMinimizeKoopmanError(Task):
         structure: Union[mtr.QChemStructure, mtr.QChemFragments, mtr.Structure],
         settings: Optional[mtr.Settings] = None,
         epsilon: Optional[Union[int, float]] = 1.0,
+        alpha: Optional[float] = None,
+        num_evals: Optional[int] = 5,
+    ) -> float:
+        def _objective(omega, _alpha):
+            beta = 1 / epsilon - _alpha
+
+            s = self.defaults(settings)
+            s["rem", "hf_sr"] = int(round(1000 * _alpha))
+            s["rem", "hf_lr"] = int(round(1000 * (_alpha + beta)))
+            s["xc_functional"] = (
+                ("X", "HF", _alpha),
+                ("X", "wPBE", beta),
+                ("X", "PBE", 1 - _alpha - beta),
+                ("C", "PBE", 1.0),
+            )
+            omega = int(round(1000 * omega))
+            s["rem", "omega"] = s["rem", "omega2"] = omega
+
+            wd = mtr.expand(f"{io.work_dir}/{omega}")
+
+            gs_io = mtr.IO("gs.in", "gs.out", wd)
+            cation_io = mtr.IO("cation.in", "cation.out", wd)
+            anion_io = mtr.IO("anion.in", "anion.out", wd)
+
+            ke = mtr.QChemKoopmanError(self.engine, gs_io, cation_io, anion_io)
+            # FIXME: not sure the best way to handle num_consumers here...
+            return ke.run(structure, s, num_consumers=3)
+
+        with self.io() as io:
+            if alpha is None:
+                return mtr.MaxLIPOTR(_objective).run(
+                    x_min=[0, 1e-3], x_max=[1 / epsilon, 1], num_evals=num_evals
+                )
+            else:
+                return mtr.MaxLIPOTR(functools.partial(_objective, _alpha=alpha)).run(
+                    x_min=1e-3, x_max=1, num_evals=num_evals
+                )
+
+
+class QChemMinimizeKoopmanErrorLPSCF(Task):
+    def __init__(
+        self,
+        engine: mtr.Engine,
+        io: mtr.IO,
+        handlers: Optional[Iterable[mtr.Handler]] = None,
+        name: Optional[str] = None,
+    ) -> None:
+        super().__init__(
+            (engine.num_threads or 1) * (engine.num_processors or 1),
+            handlers=handlers,
+            name=name,
+        )
+        self.engine = engine
+        self.io = io
+
+    def defaults(self, settings: mtr.Settings) -> mtr.Settings:
+        if ("rem", "basis") not in settings:
+            settings["rem", "basis"] = "3-21G"
+        if ("rem", "jobtype") not in settings:
+            settings["rem", "jobtype"] = "sp"
+        if ("rem", "exchange") not in settings:
+            settings["rem", "exchange"] = "gen"
+        if ("rem", "lrc_dft") not in settings:
+            settings["rem", "lrc_dft"] = True
+        if ("rem", "src_dft") not in settings:
+            settings["rem", "src_dft"] = 2
+
+        return settings
+
+    def run(
+        self,
+        *fragments: mtr.Structure,
+        active_fragment: int,
+        settings: Optional[mtr.Settings] = None,
+        epsilon: Optional[Union[int, float]] = 1.0,
         alpha: Optional[float] = 0.2,
         num_evals: Optional[int] = 5,
     ) -> float:
@@ -914,9 +1211,14 @@ class QChemMinimizeKoopmanError(Task):
                 cation_io = mtr.IO("cation.in", "cation.out", wd)
                 anion_io = mtr.IO("anion.in", "anion.out", wd)
 
-                ke = mtr.QChemKoopmanError(self.engine, gs_io, cation_io, anion_io)
+                ke = mtr.QChemKoopmanErrorLPSCF(self.engine, gs_io, cation_io, anion_io)
                 # FIXME: not sure the best way to handle num_consumers here...
-                return ke.run(structure, s, num_consumers=3)
+                return ke.run(
+                    *fragments,
+                    active_fragment=active_fragment,
+                    settings=settings,
+                    num_consumers=3,
+                )
 
             return mtr.MaxLIPOTR(f).run(x_min=1e-3, x_max=1, num_evals=num_evals)
 
@@ -931,9 +1233,7 @@ class QChemOptimize(QChemBaseTask):
         except AttributeError:
             return None
         # FIXME: is this the correct unit?
-        atoms = (
-            mtr.Atom(element=Z, position=p * mtr.angstrom) for Z, p in zip(zs, coords)
-        )
+        atoms = (mtr.Atom(element=Z, position=p) for Z, p in zip(zs, coords))
         return mtr.Structure(*atoms)
 
     def defaults(self, settings: mtr.Settings) -> mtr.Settings:
@@ -1271,3 +1571,24 @@ class WriteQChemInputSinglePoint(WriteQChemInput):
 #                 )
 #             else:
 #                 subprocess.call([self.settings["executable"]], stdout=f)
+
+
+class XTBOptimize(ExternalTask):
+    def parse(self, output: str) -> mtr.Structure:
+        with open(output, "r") as f:
+            structure_file = re.search(
+                r"optimized geometry written to:\s*(?P<structure>.*)\s*",
+                "".join(f.readlines()),
+            ).group("structure")
+
+        return mtr.Structure.read(f"{pathlib.Path(output).parent}/{structure_file}")
+
+    def run(self, structure: Union[mtr.Structure, str]) -> mtr.Structure:
+        with self.io() as io:
+            if isinstance(structure, mtr.Structure):
+                with structure.tempfile(suffix=".xyz", dir=io.work_dir) as f:
+                    self.engine.execute(f.name, self.io, arguments=["--opt"])
+            else:
+                self.engine.execute(structure, self.io, arguments=["--opt"])
+
+            return self.parse(io.out)
